#!/usr/bin/ruby
#--
# Copyright (c) 2011-2016 Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#++

require "rubygems"

require "bughunting/web/controller"
require "erb"
require "json"
require "sinatra"
require "haml"
require 'logger'
require 'puma'

enable :sessions
#enable :logging, :dump_errors, :raise_errors

############################################################

## TODO: move to config file

# Enable re-login (multiple times) for all users
MULTI_LOGIN_ALL = true

# Don't collect name and email
ANONYMOUS = true

# [CTF] No score, but a 'flag' is returned instead
FLAGS = false

# [FixCode] To not show welcome screen or the button to return to it
WELCOME = false

# Show all players (not just with non-zero score)
ALL_PLAYERS = false

# Display
TITLE = 'FixCode'
LOGO = '/img/logo-red_hat-summer-camp-brno.png'

# Show language column
SHOW_LANG = false

############################################################

include ERB::Util

controller = WebController.new "/etc/bughunting/web.yml"

# server configuration

configure do
  set :bind, controller.config["listen.ip"]
  set :port, controller.config["listen.port"]
  #set :environment, :development
  set :logger, Logger.new(STDOUT)

  set :environment, :production
  set :logging, false

  set :controller, controller
end

# helper functions

helpers do
  def local_access?
    ["127.0.0.1", "::1"].include? request.ip
  end

  def service_access?
    request.ip == settings.controller.config["check_server.ip"] \
      or request.ip == settings.controller.config["admin.ip"]
  end

  def condition_enforce_uri(condition, target_uri, return_uri = "/", error_message = nil)
    if condition and request.path != target_uri
      logger.info "Redirecting to #{target_uri}, reason: #{error_message}" if error_message
      redirect to(target_uri)
    elsif not condition and request.path == target_uri
      logger.info "You should not go to #{target_uri}, not: #{error_message}" if error_message
      redirect to(return_uri)
    else
      # true if current URI should be enforced
      condition
    end
  end

  def machine_name(ip, verbose = false)
    name = settings.controller.machine_name ip
    if verbose and name != ip
      "%s (%s)" % [name, ip]
    else
      name
    end
  end

  def registered? player
    settings.controller.players.name player
  end

  def is_ready? param
    session[:ready] = param == "1" if param
    session[:ready]
  end

  def set_message text
    session[:message] = text
  end

  def get_message
    m = session[:message]
    session.delete :message
    m
  end
end

# access filtering

before do
  @player = session[:player]
  @player_name = controller.players.name @player
  @machine = machine_name(@player)
  @ready = is_ready? params[:ready]

  session[:player_name] = @player_name
  session[:machine] = @machine

  if service_access? or local_access?
    next

  else
    logger.info "player = #{@player.inspect}" if @player
    logger.info "params = #{params.inspect}" if @params.any?
    logger.info "message = #{session[:message].inspect}" if session[:message]
  end

  registration = (@player_name.nil? or @player_name.empty?)

  next if condition_enforce_uri \
    registration, "/registration/", '/', "Codename missing"

  # The game has started, welcome is enabled and player has read
  # the instructions
  welcome = WELCOME && !( controller.game_running? and @ready )
  next if condition_enforce_uri welcome, "/welcome/"
end

# pre-game

get "/registration/" do
  haml :registration, :locals => {
    :message => get_message,
    :anonymous => ANONYMOUS
  }
end

post "/registration/" do
  if params[:terms_of_service] == "1"
    if MULTI_LOGIN_ALL
      multi = true
    end

    if params[:machine] == 'test'
      multi = true
      params[:key] = '1234567890101112'
    end

    if params[:machine] == 'devel'
      multi = true
      params[:key] = '9876543210123456'
    end

    @player = controller.players.register(
      params[:machine],
      params[:name],
      params[:email],
      params[:key],
      params[:terms_of_service],
      multi)

    if @player
      logger.info "Login succesful."
      session[:player] = @player
      redirect("/")
      return
    else
      set_message "Login failed, unknown Codename or Key."
    end
  else
    set_message "You have to agree to terms of service."
  end

  redirect("/registration/")
end

get "/welcome/" do
  haml :welcome
end

## client requests
get "/" do
  haml :index, :locals => {
    :tasks => controller.tasks_list(@player),
    :player_name => @player_name,
    :machine_name => machine_name(@player),
    :total_score => controller.scores.total(@player),
    :local_access => local_access?,
    :key => @player
  }
end

get "/task/:name/" do
  task_name = params[:name]

  task = controller.tasks[task_name]
  halt 404 if task.nil?

  haml :task, :locals => {
    :task => task,
    :score => controller.scores.task(@player, task_name),
    :revealed_hints => controller.hints.number_of_revealed(@player, task_name),
  }
end

get "/task/:name/hint" do
  task_name = params[:name]
  hint_number = params[:num].to_i

  # task existence
  task = controller.tasks[task_name]
  halt 404 if task.nil?

  # hint existence
  halt 404 if hint_number < 1 or hint_number > task.config["hints"].count

  revealed = controller.hints.number_of_revealed @player, task_name
  if revealed + 1 == hint_number
    controller.hints.reveal @player, task_name, hint_number
  end

  redirect to("/task/%s/" % task_name)
end

# results

get "/scoreboard/" do
  #halt 403 unless local_access? or service_access?

  if params.include? "json"
    ranking = controller.scores.ranking(all_players: ALL_PLAYERS)
    ranking.collect! do |rank|
      {
        :player => rank[:player],
        :name => machine_name(rank[:player]),
        :score => rank[:score],
      }
    end
    best_rank = ranking.max_by do |rank| rank[:score] end

    JSON.generate({
      "max_score" => best_rank.nil? ? 0 : best_rank[:score],
      "ranking" => ranking,
    })
  else
    haml :scoreboard, :layout => false
  end
end

get "/results/" do
  halt 403 unless service_access? or local_access?
  haml :results, :layout => false, :locals => {
    :ranking => controller.scores.ranking
  }
end

get "/logout/" do
  session.delete :player
  session.delete :ready
  redirect("/registration/")
end

#admin

get "/admin/" do
        halt 403 unless local_access?

        haml :admin, :layout => true, :locals => {
                 :running => controller.game_running?
        }
end

post "/admin/start" do
        halt 403 unless local_access?

        controller.start_game
        redirect("/admin/")
end

post "/admin/stop" do
        halt 403 unless local_access?

        controller.stop_game
        redirect("/admin/")
end

post "/admin/archive" do
        halt 403 unless local_access?

        controller.archive_game
        redirect("/admin/")
end

# service requests

post "/_service/login" do
  halt 403 unless service_access?

  player = params[:player]

  halt 404 unless registered? player
end

post "/_service/solved" do
  halt 403 unless service_access?

  # BEWARE: @player != player

  player = params[:player]
  task_name = params[:task]
  score = params[:score].to_i

  halt 404 unless registered? player
  halt 404 if task_name.nil? or task_name.empty?

  task = controller.tasks[task_name]
  halt 404 if task.nil?
  halt 404 if player.nil? or player.empty?

  current_score = controller.scores.task player, task_name
  revealed_hints = controller.hints.number_of_revealed player, task_name
  possible_max = task.config["max_score"] - revealed_hints

  if score > possible_max
    score = possible_max

  end

  if FLAGS  && score > 0
    flag = controller.flag player, task_name, score
    halt 500 if flag.nil? or flag.empty?

    msg = "TunaCTF{" + flag.to_s + "}"
    [200, msg]

  elsif score > current_score
    controller.scores.set player, task_name, score
    [200, score.to_s]

  else
    [200, current_score.to_s]

  end
end

# errors

not_found do
  haml :e404
end
